
#line 3 "template.c"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "TEMPLATE.h"
#include "TEMPLATETables.h"

static	void	initTables() ;
static	int	firstTime = 1 ;

void
transformDistribution(in, out)
  Distribution *in ;
  Distribution *out ;
{
  int		meanIdx, varianceIdx ;
  double	tmp1, tmp2 ;
  double	deltaMean, deltaVariance, meanRatio, varianceRatio ;
  Distribution	*bound[2][2] ;
  double	logVariance = log(in->variance) ;

  if (firstTime) {
    initTables() ;
    firstTime = 0 ;
  }
  
  deltaMean	= (maxMean - minMean)/(numMeanSamples - 1) ;
  deltaVariance = (maxVariance - minVariance)/(numVarianceSamples -1) ;

  if (in->mean <= minMean)
    meanIdx = 1 ;
  else if (in->mean >= maxMean)
    meanIdx = numMeanSamples - 1 ;
  else 
    meanIdx = (in->mean - minMean)/deltaMean + 1 ;

  if (logVariance <= minVariance)
    varianceIdx = 1 ;
  else if (logVariance >= maxVariance)
    varianceIdx = numVarianceSamples - 1 ;
  else 
    varianceIdx = (logVariance - minVariance)/deltaVariance + 1 ;

  bound[0][0] = &sigmoidTable[meanIdx - 1][varianceIdx - 1] ;
  bound[0][1] = &sigmoidTable[meanIdx - 1][varianceIdx] ;
  bound[1][0] = &sigmoidTable[meanIdx][varianceIdx - 1] ;
  bound[1][1] = &sigmoidTable[meanIdx][varianceIdx] ;

  meanRatio
    = (in->mean - (minMean + (meanIdx - 1)*deltaMean))/deltaMean ;
  varianceRatio
    = (in->variance - exp(minVariance + deltaVariance*(varianceIdx - 1)))
      /(  exp(minVariance + deltaVariance*varianceIdx)
	- exp(minVariance + deltaVariance*(varianceIdx - 1))) ;

  tmp1 = bound[0][0]->mean
    + (bound[0][1]->mean - bound[0][0]->mean)*varianceRatio ;
  tmp2 = bound[1][0]->mean
    + (bound[1][1]->mean - bound[1][0]->mean)*varianceRatio ;
  out->mean = tmp1 + (tmp2 - tmp1)*meanRatio ;

  tmp1 = bound[0][0]->variance
    + (bound[0][1]->variance - bound[0][0]->variance)*varianceRatio ;
  tmp2 = bound[1][0]->variance 
    + (bound[1][1]->variance - bound[1][0]->variance)*varianceRatio ;
  out->variance = tmp1 + (tmp2 - tmp1)*meanRatio ;

  tmp1 = bound[0][0]->dMean_dVariance
    + (bound[0][1]->dMean_dVariance
       - bound[0][0]->dMean_dVariance)*varianceRatio ;
  tmp2 = bound[1][0]->dMean_dVariance 
    + (bound[1][1]->dMean_dVariance
       - bound[1][0]->dMean_dVariance)*varianceRatio ;
  out->dMean_dVariance = tmp1 + (tmp2 - tmp1)*meanRatio ;

  tmp1 = bound[0][0]->dMean_dMean
    + (bound[0][1]->dMean_dMean 
       - bound[0][0]->dMean_dMean)*varianceRatio ;
  tmp2 = bound[1][0]->dMean_dMean 
    + (bound[1][1]->dMean_dMean
       - bound[1][0]->dMean_dMean)*varianceRatio ;
  out->dMean_dMean = tmp1 + (tmp2 - tmp1)*meanRatio ;

  tmp1 = bound[0][0]->dVariance_dVariance
    + (bound[0][1]->dVariance_dVariance
       - bound[0][0]->dVariance_dVariance)*varianceRatio ;
  tmp2 = bound[1][0]->dVariance_dVariance 
    + (bound[1][1]->dVariance_dVariance
       - bound[1][0]->dVariance_dVariance)*varianceRatio ;
  out->dVariance_dVariance = tmp1 + (tmp2 - tmp1)*meanRatio ;

  tmp1 = bound[0][0]->dVariance_dMean
    + (bound[0][1]->dVariance_dMean
       - bound[0][0]->dVariance_dMean)*varianceRatio ;
  tmp2 = bound[1][0]->dVariance_dMean 
    + (bound[1][1]->dVariance_dMean
       - bound[1][0]->dVariance_dMean)*varianceRatio ;
  out->dVariance_dMean = tmp1 + (tmp2 - tmp1)*meanRatio ;
}

#ifdef SLOW
static void
initTables() {
  FILE	*fp ;
  int	row, column ;
  char	*dataFile ;

  dataFile = getenv("TEMPLATEDATA") ;

  if (dataFile == NULL) {
    sprintf(path, "%s/lib/TEMPLATESigmoid.dat", getenv("XERIONDIR")) ;
    dataFile = path ;

    fprintf(stderr, "Loading default table:  %s...", dataFile) ;
  } else {
    fprintf(stderr, "Loading table: %s...", dataFile) ;
  }

  fp = fopen(dataFile, "r") ;

  for (row = 0 ; row < numMeanSamples ; ++row) {
    for (column = 0 ; column < numVarianceSamples ; ++column) {
      Distribution	*item = &sigmoidTable[row][column] ;
      if (fscanf(fp, "%lg %lg %lg %lg %lg %lg",
		 &item->mean, &item->variance, 
		 &item->dMean_dVariance, &item->dMean_dMean,
		 &item->dVariance_dVariance, &item->dVariance_dMean) != 6) {
	fprintf(stderr, "Error initializing data table!!!\n") ;
	exit(1) ;
      }
    }
  }
  fclose(fp) ;
  fprintf(stderr, "done\n") ;
}
#else
static	void 
initTables(void) {
  int	fd ;
  int	row, column ;
  char	path[BUFSIZ] ;
  char	*dataFile ;

  dataFile = getenv("TEMPLATEDATA") ;

  if (dataFile == NULL) {
    sprintf(path, "%s/lib/TEMPLATESigmoid.dat", getenv("XERIONDIR")) ;
    dataFile = path ;

    fprintf(stderr, "Loading default table:  %s...", dataFile) ;
  } else {
    fprintf(stderr, "Loading table: %s...", dataFile) ;
  }

  fd = open(dataFile, O_RDONLY) ;
  if (fd == -1) {
    fprintf(stderr, "Cannot open data file!!!\n") ;
    exit(1) ;
  }
  for (row = 0 ; row < numMeanSamples ; ++row) {
    for (column = 0 ; column < numVarianceSamples ; ++column) {
      Distribution	*item = &sigmoidTable[row][column] ;
      double		tmp[6] ;
      if (!read(fd, tmp, sizeof(tmp))) 
	fprintf(stderr, "Error initializing data table!!!\n"), exit(1) ;
      item->mean 		= tmp[0] ;
      item->variance 		= tmp[1] ;
      item->dMean_dVariance	= tmp[2] ;
      item->dMean_dMean		= tmp[3] ;
      item->dVariance_dVariance = tmp[4] ;
      item->dVariance_dMean	= tmp[5] ;
    }
  }
  close (fd) ;
  fprintf(stderr, "done\n") ;
}
#endif
